subroutine findequil(err) !Determines model equilibrium taking a number of global variables as given and returning equilibrium in global variables.
  !If solution error occurs, the rotuine returns with err=1.
  use modelvar
  use mpivar
  implicit none
  integer, intent(out) :: err
  integer :: i,j
  real(8) :: psi_1(en),delta_1,distance,g_1
  real(8), parameter :: tol=1.0e-9
  real(16) :: x,xx,xxx
  psi_1=zero !Initial guess on Psi
  delta_1=0.2 !Initial guess on delta
  g_1=0.015 !Initial guess on growth
  distance=one
  i=1
  do while (distance .gt. tol .and. i .lt. 1000) !iteration loop to determine fixed point. Stop once distance between iterations below tolerance.
     psi_v=psi_1
     delta=delta_1
     g=g_1
     call parcalc(err) !(psi,delta,g) mapping. Returns in psi_1,delta_1 and g_1.
     distance=max(maxval(abs(psi_v-psi_1)),abs(delta-delta_1),abs(g-g_1)) !metric is simple max of absolute differences
     i=i+1
  end do
  if(i==1000 .and. distance .gt. tol) then
     err=1
     print *,'Equilibrium solver did not achieve desired precision.'
     return
  end if
  psi_v=psi_1
  delta=delta_1
  g=g_1
  call parcalc(err) !global variables set with equilibrium solution and final run with parcalc is done to test error status.
  if(err .eq. 1) RETURN
  do i=1,en !Calculation of type dependent steady state mass of firms.
     if(gam(i) .gt. zero) then
        capem(i)=log(delta/(delta-gam(i)))*eta*phi(i)/gam(i)
     else
        capem(i)=eta*phi(i)/delta
     end if
  end do
  do i=1,en !Calculation of type dependent steady state size distribution.
     x=zero
     xx=log(delta/(delta-gam(i)))
     xxx=gam(i)/delta
     do j=1,kbar
        x=x+(xxx**dble(j))/(dble(j)*xx)
        sizecdf(j,i)=dble(x)
     end do
  end do
contains
  subroutine parcalc(err)
    implicit none
    integer, intent(out) :: err
    real(8) :: l1,l2,l3,l4,reslt(en),fnorm
    integer :: i
    err=0
    do i=1,en
       v(i)=pi(i)/(r+delta-g*(one-sigma)) + psi_v(i) !type dependent expected value of an additional product.
       if(v(i) > zero) then !determination of optimal type dependent creation rates.
          gam(i)=gamfonc(v(i))
       else
          gam(i)=zero
       end if
    end do
    call ksolver(err) !provides a solution to k for given eta and gam (type dependent steady state mass of products).
    if(err == 1) RETURN
    delta_1=zero
    do i=1,en
       delta_1=delta_1+k(i)*gam(i)
    end do
    delta_1=delta_1+eta !update delta
    g_1=g_calc() !update g
    g_1=min(delta_1,g_1) !This line is inserted to ensure that the solver does not crash. In fixed point, check if this is binding - it must not be.
    do i=1,en !update psi (accelerated by use of delta_1 rather than delta... can also destabilize iteration).
       psi_1(i)=(gam(i)*v(i)-w*c(gam(i)))/(r+delta_1)
    end do
  end subroutine parcalc
end subroutine findequil

subroutine ksolver(err) !solves for the steady state k-distribution. Takes input from global variables: eta, gam, phi. Gives output in k.
  use modelvar
  use mpivar
  implicit none
  integer, intent(out) :: err
  real(8) :: k0(en), k1(en),distance,rparam(7)
  real(8), parameter :: tol=1.0e-9
  real(8), dimension(en), parameter :: lobound=zero,hibound=one,xscale=one
  real(8), parameter :: fscale=one
  integer :: i,j,iparam(7)
  err=0
  k0=phi !Initial guess
  call erset(0,1,0)
  call du4inf(iparam,rparam)
  iparam(3)=10000
  iparam(4)=40000
  iparam(5)=40000
  call dbcong(kfcn1,jacob1,en,k0,0,lobound,hibound,xscale,fscale,iparam,rparam,k0,distance) !call to simple IMSL minimization routine to find fixed point of k-mapping.
  if(distance .gt. dble(1.0e-6)) then !check if fixed point is found.
     err=1
     if(master) then
        write(6,'("Warning: fixed point not found precisely.")')
        print *,distance
        print *,k0
        print *,gam
        print *,delta
     end if
     stop
  end if
  call erset(0,2,2)
  k=k0/sum(k0) !outputs in global variable.
  if(abs(sum(k0)-one) .gt. 1.0e-5) then
     err=1
     if(master) then
        write(6,'("Warning: k-solution did not add to one")')
        print *,gam
        print *,eta,sum(k0)
     end if
  end if
contains
  subroutine kmap(inp,outp) !k-mapping
    implicit none
    real(8), intent(in) :: inp(en)
    real(8), intent(out) :: outp(en)
    integer :: ii,maxfcn
    real(8) :: l_sum,lo,hi
    l_sum=zero
    do ii=1,en
       l_sum=l_sum+gam(ii)*inp(ii)
    end do
    do ii=1,en
       x1=phi(ii)
       x2=gam(ii)
       x3=l_sum-gam(ii)*inp(ii)
       lo=zero
       hi=one
       maxfcn=1000
       call dzbren(map,dble(1.0e-10),dble(1.0e-10),lo,hi,maxfcn) !simple one-dimensional minimizer bounded below and above.
       outp(ii)=hi
    end do
  end subroutine kmap

  double precision function map(inp)
    implicit none
    real(8), intent(in) :: inp
    map=inp-(eta*x1+inp*(one-inp)*x2)/(eta+x3)
  end function map

  subroutine kfcn1(dim,inp,outp) !distance between input and output of k-map.
    implicit none
    integer, intent(in) :: dim
    real(8), intent(in) :: inp(dim)
    real(8), intent(out) :: outp
    real(8) :: om(dim),loc_inp(dim)
    integer :: i
    loc_inp=inp
    call kmap(loc_inp,om)
    outp=zero
    do i=1,en
       outp=outp+(om(i)-loc_inp(i))**2
    end do
  end subroutine kfcn1

  subroutine jacob1(dim,inp,outp) !jacobian for kfcn1
    implicit none
    integer, intent(in) :: dim
    real(8), intent(in) :: inp(dim)
    real(8), intent(out) :: outp(dim)
    real(8) :: mat(dim,dim)
    real(8) :: om(dim),loc_sum,loc_inp(dim),grad
    integer :: ii,jj
    loc_inp=inp
    call kmap(loc_inp,om)
    do jj=1,dim
       loc_sum=zero
       do ii=1,dim
          if(ii .ne. jj) loc_sum=loc_sum+gam(ii)*loc_inp(ii)
       end do
       grad=-2.0d0*(om(jj)-loc_inp(jj))
       do ii=1,dim
          if(ii .eq. jj) then
             mat(jj,ii)=grad
          else
             mat(jj,ii)=gam(ii)*om(jj)*grad/(eta+loc_sum-gam(jj)*(one-dble(2)*om(jj)))
          end if
       end do
    end do
    outp=sum(mat,dim=1)
  end subroutine jacob1
end subroutine ksolver

subroutine simul(yy,ww,nn,setseed) !Requires that firmequil routine has been called prior to call.
  use modelvar
  use distr
  use mpivar
  implicit none
  integer(4), intent(in) :: setseed
  real(4), intent(out) :: yy(n,t),ww(n,t),nn(n,t)
  real(8) :: uni(n),portfolio(kbar+1,2,t),hazgam,hazdel,uni1(2*per),l1(per),l2(n),l3(1)
  integer :: ini(n,2),i,type,j,l0,ii,currk,kissflag,iii
  integer(4) :: cdseed(per,2),seed1,seed2(n),seed3(n,t-1)
  logical :: flag
  call rnset(setseed) !set seed for simulation according to input setseed.

  seed1=int(drnunf()*seedmult+1)
  call drnun(n,l2)
  seed2=int(l2*seedmult+1)
  do i=1,t-1
     call drnun(n,l2)
     seed3(:,i)=int(l2*seedmult+1)
  end do

  yy=zero
  ww=zero
  nn=zero

  !Start out firms in steady state

  call rnset(seed1) !seed is set to stabilize firm size and type simulation across different model parameter choices.
  call drnun(n,uni)

  call ergodist(uni,ini)

  do i=1,n
     portfolio=zero
     !first put together the initial product portfolio.
     currk=ini(i,2)
     type=ini(i,1)

     call rnset(seed2(i)) !stabilize firm i's type zero product draws

     do j=1,currk
        portfolio(j,1,1)=drawq(type)
        portfolio(j,2,1)=drawz()*exp(g*(one-sigma)*age()) !initial demand is drawn from drawz() and then the age of the product is drawn from age().
     end do !done

     !then, begin to simulate forward
     call update !provides up to date values for hazards.
     do j=2,T
        if(currk .eq. 0) exit
        portfolio(:,1,j)=portfolio(:,1,j-1)
        portfolio(:,2,j)=portfolio(:,2,j-1)*exp(g*(one-sigma))
        call rnset(seed3(i,j-1))
        call drnun(2*per,uni1)

	do ii=1,2 !generate cdseed to be used in destroy and create calls
           call drnun(per,l1)
           cdseed(:,ii)=int(l1*seedmult+1)
        end do
        do ii=1,per
           if(currk .eq. 0) exit
           if(uni1((ii-1)*2+1) .lt. hazdel) call destroy
           if(uni1((ii-1)*2+2) .lt. hazgam) call create
           if(flag) call update
        end do
     end do

     do j=1,T
        currk=0
        ii=1
        do while(portfolio(ii,1,j) .ne. zero)
           yy(i,j)=yy(i,j)+portfolio(ii,2,j)
           ww(i,j)=ww(i,j)+w*portfolio(ii,2,j)/(markup(portfolio(ii,1,j))*(w+kappa))
           ii=ii+1
           currk=currk+1
        end do
        ww(i,j)=ww(i,j)+w*dble(currk)*Z*c(gam(type))
     end do
  end do

  do i=1,N !measurement error addition
     do j=1,T
        call drnnoa(1,l3(1))
        if(yy(i,j) .gt. dble(0)) yy(i,j)=exp(min(varhuge,log(yy(i,j))+l3(1)*sqrt(errorsigmaY)-errorsigmaY/2.0d0))
        call drnnoa(1,l3(1))
        if(ww(i,j) .gt. dble(0)) ww(i,j)=exp(min(varhuge,log(ww(i,j))+l3(1)*sqrt(errorsigmaW)-errorsigmaW/2.0d0))
     end do
  end do
  nn=ww/w


  if(.not. kill_g) then
     do i=2,T !deflate y and w according to final consumption good price.
        yy(:,i)=yy(:,i)*exp(g*dble(i))
        ww(:,i)=ww(:,i)*exp(g*dble(i))
     end do
  end if
contains
 subroutine destroy
    implicit none
    integer :: l1,iii
    call rnset(cdseed(ii,1)) !stabilize the destruction draw

    l1=1+drnunf()*dble(currk)
    do iii=l1,currk
       portfolio(iii,1:2,j)=portfolio(iii+1,1:2,j)
    end do
    currk=currk-1
    flag=.TRUE.
  end subroutine destroy

  subroutine create
    implicit none
    call rnset(cdseed(ii,2)) !stabilize creation draws.

    currk=currk+1
    portfolio(currk,1,j)=drawq(type)
    portfolio(currk,2,j)=drawz()
    flag=.TRUE.
  end subroutine create

  subroutine update
    implicit none
    if(currk .lt. kbar) then !check whether # of products is below max
       hazgam=one-exp(-dble(currk)*gam(type)/dble(per))
    else
       hazgam=zero
    end if
    hazdel=one-exp(-dble(currk)*delta/dble(per))
    flag=.FALSE.
  end subroutine update

  double precision function drawq(inp)
    implicit none
    integer, intent(in) :: inp
    if(kill_q_var) then
       drawq=one/(one-pi(inp)) !use for zero q variance counterfactual in FHK table
    else
       drawq=weibull_invcdf(drnunf(),q_beta,sig(inp))+one
    end if
  end function drawq

  double precision function drawz()
    implicit none
    drawz=weibull_invcdf(drnunf(),Z_beta,Z_eta)+Z_gam
  end function drawz

  double precision function age()
    implicit none
    if(kill_a_var) then
       age=zero !counterfactual to eliminate age effects
    else
       age=( log(delta)-log(drnunf()) )/delta
    end if
  end function age
end subroutine simul

subroutine entry_simul(setseed) !Requires that esteval routine has been called prior to call - besides finding equilibrium, em must have been determined.
  !This routine simulates a set of entrants for given equilibrium parameters. Unlike the simul routine, the number of entrants over a given period is a random variable.
  !Therefore, it is not known at the time of the call of entry_simul how many firm observations will be given at output. As a result, output will be given in global
  !variables, yy_e,ww_e,nn_e which are defined as allocatable arrays.
  use modelvar
  use distr
  use mpivar
  implicit none
  integer(4), intent(in) :: setseed
  real(8) :: portfolio(kbar+1,2,t),hazgam,hazdel,uni1(2*per),l1(per),l3(1)
  integer :: ini(n,2),i,type,j,l0,ii,currk
  integer(4) :: cdseed(per,2),seed1
  logical :: flag
  integer :: N_e !number of potential entrants
  real(8), allocatable :: uni_e(:),uni(:),l2(:)
  integer(4), allocatable :: ini_e(:),seed2(:),seed3(:,:)
  integer(2), allocatable :: set_e(:),ini_type(:)
  integer :: flag_e,start_per
  real(8) :: hazentr !entry hazard
  integer(4) :: N1 !number of actual entrants... may exceed the final size of yy_e because some entrants will enter and exit within same year in which case entrant is deleted.

  N_e=dble(N)*em/sum(capem) !Number of potential entrants.
  game=eta/em !individual entrant's entry rate.
  allocate(uni_e(N_e),set_e(per*(T-1)))

  call rnset(setseed)
  seed1=int(drnunf()*seedmult+1)

  !First, determine the full set of entrants. To begin with, simply register the entry period of entrant.

  set_e=0
  N1=0
  hazentr=one-exp(-game/dble(per))
  call rnset(seed1)

  do i=1,per*(T-1)
     call drnun(N_e,uni_e)
     do j=1,N_e
        if(uni_e(j) .lt. hazentr) then
           set_e(i)=set_e(i)+1 !set_e(i) is the number of entrants in period i
           N1=N1+1
        end if
     end do
  end do

  allocate(ini_e(N1))
  ii=0
  do i=1,per*(T-1)
     do j=1,set_e(i)
        ii=ii+1
        ini_e(ii)=i !ini_e(ii) is the entry period for the ii'th entrant
     end do
  end do
  deallocate(set_e)

  if(allocated(yy_e)) deallocate(yy_e,ww_e,nn_e)
  allocate(yy_e(N1,T),ww_e(N1,T),nn_e(N1,T))

  allocate(uni(N1),ini_type(N1),seed2(N1),seed3(N1,T-1),l2(N1))

  yy_e=zero
  ww_e=zero
  nn_e=zero

  call drnun(N1,l2)
  seed2=int(l2*seedmult+1)
  do i=1,t-1
     call drnun(N1,l2)
     seed3(:,i)=int(l2*seedmult+1)
  end do
  
  call drnun(N1,uni)
  do i=1,N1
     do j=1,en
        if(uni(i) .lt. sum(phi(1:j))) then
           ini_type(i)=j !The type of entrant i
           exit
        end if
     end do
  end do

  do i=1,N1 !loop over set of entrants
     type=ini_type(i)
     portfolio=zero
     currk=0
     flag_e=1
     do j=2,T !loop over years
        if(flag_e==1) then !if entrant has yet to enter, determine if entry in this year
           if((ini_e(i)-1)/per > j-2) cycle !if entry period is not in the current year then cycle
           call rnset(seed2(i))
           currk=1
           portfolio(currk,1,j)=drawq(type)
           portfolio(currk,2,j)=drawz() !demand realization at age 0.
           call update
           flag_e=0
           start_per=ini_e(i)-per*(j-2)
        else !in the case where entrant is already in. In this case age the portfolio and move on to standard dynamics.
           start_per=1
           portfolio(:,:,j)=portfolio(:,:,j-1)*exp(g*(one-sigma))
        end if
        if(currk .eq. 0) exit
        call rnset(seed3(i,j-1))
        call drnun(2*per,uni1)
	do ii=1,2
           call drnun(per,l1)
           cdseed(:,ii)=int(l1*seedmult+1)
        end do
        do ii=start_per,per
           if(currk .eq. 0) exit
           if(uni1((ii-1)*2+1) .lt. hazdel) call destroy
           if(uni1((ii-1)*2+2) .lt. hazgam) call create
           if(flag) call update
        end do
     end do

     do j=1,T
        currk=0
        ii=1
        do while(portfolio(ii,1,j) .ne. zero)
           yy_e(i,j)=yy_e(i,j)+portfolio(ii,2,j)
           ww_e(i,j)=ww_e(i,j)+w*portfolio(ii,2,j)/(markup(portfolio(ii,1,j))*(w+kappa))
           ii=ii+1
           currk=currk+1
        end do
        ww_e(i,j)=ww_e(i,j)+w*dble(currk)*Z*c(gam(type))
     end do
  end do

  do i=1,N1 !measurement error addition
     do j=1,T
        call drnnoa(1,l3(1))
        if(yy_e(i,j) .gt. dble(0)) yy_e(i,j)=exp(min(varhuge,log(yy_e(i,j))+l3(1)*sqrt(errorsigmaY)-errorsigmaY/2.0d0))
        call drnnoa(1,l3(1))
        if(ww_e(i,j) .gt. dble(0)) ww_e(i,j)=exp(min(varhuge,log(ww_e(i,j))+l3(1)*sqrt(errorsigmaW)-errorsigmaW/2.0d0))
     end do
  end do
  nn_e=ww_e/w

  if(.not. kill_g) then
     do i=2,T !deflate y and w according to final consumption good price.
        yy_e(:,i)=yy_e(:,i)*exp(g*dble(i))
        ww_e(:,i)=ww_e(:,i)*exp(g*dble(i))
     end do
  end if

  deallocate(uni,ini_type,seed2,seed3,l2)
contains
  subroutine destroy
    implicit none
    integer :: l1,iii
    call rnset(cdseed(ii,1)) !stabilize the destruction draw
    l1=1+drnunf()*dble(currk)
    do iii=l1,currk
       portfolio(iii,1:2,j)=portfolio(iii+1,1:2,j)
    end do
    currk=currk-1
    flag=.TRUE.
  end subroutine destroy

  subroutine create
    implicit none
    call rnset(cdseed(ii,2)) !stabilize creation draws.
    currk=currk+1
    portfolio(currk,1,j)=drawq(type)
    portfolio(currk,2,j)=drawz()
    flag=.TRUE.
  end subroutine create

  subroutine update
    implicit none
    if(currk .lt. kbar) then !check whether # of products is below max
       hazgam=one-exp(-dble(currk)*gam(type)/dble(per))
    else
       hazgam=zero
    end if
    hazdel=one-exp(-dble(currk)*delta/dble(per))
    flag=.FALSE.
  end subroutine update

  double precision function drawq(inp)
    implicit none
    integer, intent(in) :: inp
    if(kill_q_var) then
       drawq=one/(one-pi(inp)) !use for zero q variance counter factual in FHK table
    else
       drawq=weibull_invcdf(drnunf(),q_beta,sig(inp))+one
    end if
  end function drawq

  double precision function drawz()
    implicit none
    drawz=weibull_invcdf(drnunf(),Z_beta,Z_eta)+Z_gam
  end function drawz
end subroutine entry_simul
